#!/usr/bin/env python
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2011-2025 Jarkko Sakkinen <jarkko.sakkinen@iki.fi>
# Copyright (c) 2011 Ferrix Hovi <ferrix@ferrix.fi>

import logging
import os
import re
import shutil
import stat
import sys
from datetime import datetime, timezone
from hashlib import md5
from optparse import OptionParser
from struct import unpack_from

from PIL import Image
from PIL.ExifTags import TAGS

TIFF_BIG_ENDIAN = 0x4D4D  # 'MM'
TIFF_LITTLE_ENDIAN = 0x4949  # 'II'
TIFF_DATETIME_TAG = 0x0132  # DateTime tag

TIFF_TAG_TYPE_ASCII = 2
TIFF_DATETIME_LENGTH = 20
MIN_ARGS = 2


def get_cr2_datetime(file):
    with open(file, "rb") as f:
        buffer = f.read(1024)
        byte_order = unpack_from("HHLHBBL", buffer)[0]
        endian_flag = "@"
        if byte_order == TIFF_BIG_ENDIAN:
            endian_flag = ">"
        elif byte_order == TIFF_LITTLE_ENDIAN:
            endian_flag = "<"

        datetime_offset = None
        entries = unpack_from(endian_flag + "H", buffer, 0x10)[0]
        for entry in range(entries):
            (tag_id, tag_type, num_of_value, value) = unpack_from(endian_flag + "HHLL", buffer, 0x10 + 2 + entry * 12)
            if tag_id == TIFF_DATETIME_TAG:
                if tag_type != TIFF_TAG_TYPE_ASCII or num_of_value != TIFF_DATETIME_LENGTH:
                    # Skip this tag if unexpected format
                    continue
                datetime_offset = value

        if not datetime_offset:
            return None

        exif_timestamp = "".join(unpack_from(TIFF_DATETIME_LENGTH * "s", buffer, datetime_offset)).strip("\x00")

        return datetime.strptime(exif_timestamp, "%Y:%m:%d %H:%M:%S").replace(tzinfo=timezone.utc)


def main():
    logging.basicConfig(level=logging.ERROR)
    parser = OptionParser(usage="Usage: %prog [options] <source path> <dest path>")
    (options, args) = parser.parse_args()

    if len(args) < MIN_ARGS:
        parser.print_help()
        sys.exit(1)

    source_dir_path = args[0]
    dest_dir_path = args[1]

    for source_basename in os.listdir(source_dir_path):
        source_path = os.path.join(source_dir_path, source_basename)
        source_path = os.path.abspath(source_path)

        mode = os.stat(source_path)[stat.ST_MODE]
        if not stat.S_ISREG(mode):
            continue

        try:
            with open(source_path, "rb") as im_file:
                source_hash = md5()
                while True:
                    data = im_file.read(8192)
                    if not data:
                        break
                    source_hash.update(data)
                hashsum = source_hash.hexdigest()
        except OSError:
            logging.exception("Error: %s", source_path)
            continue

        timestamp = None
        if source_path.lower().endswith(".cr2"):
            exif_dt = get_cr2_datetime(source_path)
            if exif_dt is not None:
                timestamp = exif_dt.strftime("%Y%m%d%H%M%S")
        else:
            try:
                im = Image.open(source_path)
                exif = im.getexif()
                if not exif:
                    continue
                for key, value in exif.items():
                    tag = TAGS.get(key, key)
                    if isinstance(tag, str) and tag in ["DateTimeOriginal", "DateTime"]:
                        timestamp = re.sub(r"[: ]", "", value)
            except OSError:
                logging.exception("Error: %s", source_path)
                continue

        if timestamp is None:
            continue

        fext = os.path.splitext(source_basename)[1]
        dest_basename = f"{timestamp}_{hashsum}{fext}"
        dest_path = os.path.join(dest_dir_path, dest_basename)
        dest_path = os.path.abspath(dest_path)

        shutil.copy(source_path, dest_path)


if __name__ == "__main__":
    main()
